@@ -6,6 +6,7 @@ use log::info;
 use rand::{CryptoRng, Rng};
 use serde::{Deserialize, Serialize};
 
+use crate::consensus::{Consensus, VoteResponse};
 use crate::vote::{Ballot, Proposition, SignedVote, Vote};
 use crate::{Error, Result};
 
@@ -13,16 +14,14 @@ const SOFT_MAX_MEMBERS: usize = 7;
 pub type Generation = u64;
 
 #[derive(Debug)]
-pub struct State<T: Proposition> {
-    pub elders: BTreeSet<PublicKeyShare>,
+pub struct Membership<T: Proposition> {
+    pub consensus: Consensus<Reconfig<T>>,
     // TODO: we need faulty elder detection
-    // pub faulty_elders: BTreeMap<PublicKeyShare, BTreeSet<SignedVote<T>>>,
-    pub secret_key: SecretKeyShare,
+    // pub faulty_elders: BTreeMap<PublicKeyShare, BTreeSet<SignedVote<Reconfig<T>>>>,
     pub gen: Generation,
     pub pending_gen: Generation,
     pub forced_reconfigs: BTreeMap<Generation, BTreeSet<Reconfig<T>>>, // TODO: change to bootstrap members
-    pub history: BTreeMap<Generation, SignedVote<T>>, // for onboarding new procs, the vote proving super majority
-    pub votes: BTreeMap<PublicKeyShare, SignedVote<T>>,
+    pub history: BTreeMap<Generation, SignedVote<Reconfig<T>>>, // for onboarding new procs, the vote proving super majority
 }
 
 #[derive(Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Serialize, Deserialize)]
@@ -49,28 +48,24 @@ impl<T: Proposition> Reconfig<T> {
     }
 }
 
-impl<T: Proposition> State<T> {
+impl<T: Proposition> Membership<T> {
     pub fn from(secret_key: SecretKeyShare, elders: BTreeSet<PublicKeyShare>) -> Self {
-        State {
-            elders,
-            secret_key,
+        Membership::<T> {
+            consensus: Consensus::<Reconfig<T>>::from(secret_key, elders),
             gen: 0,
             pending_gen: 0,
             forced_reconfigs: Default::default(),
             history: Default::default(),
-            votes: Default::default(),
         }
     }
 
-    pub fn random(mut rng: impl Rng + CryptoRng) -> Self {
-        State {
-            elders: Default::default(),
-            secret_key: rng.gen(),
+    pub fn random(rng: impl Rng + CryptoRng) -> Self {
+        Membership::<T> {
+            consensus: Consensus::<Reconfig<T>>::random(rng),
             gen: 0,
             pending_gen: 0,
             forced_reconfigs: Default::default(),
             history: Default::default(),
-            votes: Default::default(),
         }
     }
 
@@ -131,7 +126,7 @@ impl<T: Proposition> State<T> {
         Err(Error::InvalidGeneration(gen))
     }
 
-    pub fn propose(&mut self, reconfig: Reconfig<T>) -> Result<SignedVote<T>> {
+    pub fn propose(&mut self, reconfig: Reconfig<T>) -> Result<SignedVote<Reconfig<T>>> {
         let vote = Vote {
             gen: self.gen + 1,
             ballot: Ballot::Propose(reconfig),
@@ -141,7 +136,7 @@ impl<T: Proposition> State<T> {
         Ok(self.cast_vote(signed_vote))
     }
 
-    pub fn anti_entropy(&self, from_gen: Generation) -> Vec<SignedVote<T>> {
+    pub fn anti_entropy(&self, from_gen: Generation) -> Vec<SignedVote<Reconfig<T>>> {
         info!("[MBR] anti-entropy from gen {}", from_gen);
 
         let mut msgs = Vec::from_iter(
@@ -152,12 +147,12 @@ impl<T: Proposition> State<T> {
         );
 
         // include the current in-progres votes as well.
-        msgs.extend(self.votes.values().cloned());
+        msgs.extend(self.consensus.votes.values().cloned());
 
         msgs
     }
 
-    fn resolve_votes(&self, votes: &BTreeSet<SignedVote<T>>) -> BTreeSet<Reconfig<T>> {
+    fn resolve_votes(&self, votes: &BTreeSet<SignedVote<Reconfig<T>>>) -> BTreeSet<Reconfig<T>> {
         let (winning_reconfigs, _) = self
             .count_votes(votes)
             .into_iter()
@@ -167,6 +162,75 @@ impl<T: Proposition> State<T> {
         winning_reconfigs
     }
 
+    pub fn public_key_share(&self) -> PublicKeyShare {
+        self.consensus.public_key_share()
+    }
+
+    pub fn handle_signed_vote(
+        &mut self,
+        signed_vote: SignedVote<Reconfig<T>>,
+    ) -> Result<Option<SignedVote<Reconfig<T>>>> {
+        self.validate_signed_vote(&signed_vote)?;
+        self.log_signed_vote(&signed_vote);
+
+        let vote_response = self
+            .consensus
+            .handle_signed_vote(signed_vote, self.pending_gen)?;
+
+        match vote_response {
+            VoteResponse::Broadcast(vote) => {
+                self.pending_gen = vote.vote.gen;
+                Ok(Some(vote))
+            }
+            VoteResponse::Decided(vote) => {
+                self.history.insert(self.pending_gen, vote);
+                self.gen = self.pending_gen;
+                // clear our pending votes
+                self.consensus.votes = Default::default();
+                Ok(None)
+            }
+            VoteResponse::WaitingForMoreVotes => Ok(None),
+        }
+    }
+
+    pub fn sign_vote(&self, vote: Vote<Reconfig<T>>) -> Result<SignedVote<Reconfig<T>>> {
+        self.consensus.sign_vote(vote)
+    }
+
+    pub fn cast_vote(&mut self, signed_vote: SignedVote<Reconfig<T>>) -> SignedVote<Reconfig<T>> {
+        self.log_signed_vote(&signed_vote);
+        signed_vote
+    }
+
+    fn log_signed_vote(&mut self, signed_vote: &SignedVote<Reconfig<T>>) {
+        self.pending_gen = signed_vote.vote.gen;
+        self.consensus.log_signed_vote(signed_vote);
+    }
+
+    pub fn count_votes(
+        &self,
+        votes: &BTreeSet<SignedVote<Reconfig<T>>>,
+    ) -> BTreeMap<BTreeSet<Reconfig<T>>, usize> {
+        self.consensus.count_votes(votes)
+    }
+
+    pub fn validate_signed_vote(&self, signed_vote: &SignedVote<Reconfig<T>>) -> Result<()> {
+        if signed_vote.vote.gen != self.gen + 1 {
+            return Err(Error::VoteNotForNextGeneration {
+                vote_gen: signed_vote.vote.gen,
+                gen: self.gen,
+                pending_gen: self.pending_gen,
+            });
+        }
+
+        signed_vote
+            .proposals()
+            .into_iter()
+            .try_for_each(|(_signer, reconfig)| self.validate_reconfig(reconfig))?;
+
+        self.consensus.validate_signed_vote(signed_vote)
+    }
+
     pub fn validate_reconfig(&self, reconfig: Reconfig<T>) -> Result<()> {
         let members = self.members(self.gen)?;
         match reconfig {
